#!/usr/bin/perl

#/////////////////////////////////////////////////////////////////////////////
use constant { ROLE => 11, NAME => 12, DESCRIPTION => 13, NOT_IN_TREE => 14, ROLE_NONE => 15, ROLE_PRESENTATION => 16 };

=head2 populate_test_template
 
 Parameters  : template, test_description, testable_statement
 Returns     : string
 Description : Populates the template with the test description and testable statement 
 			   and returns the completed test string.
 
=cut

sub populate_test_template ($$$) {
	my $template = $_[0];

	# test for needed namespaces
	$xlink = index( $_[2], 'xlink' ) > -1;
	$html  = index( $_[2], '<html:' ) > -1;

	# Only add namespaces when needed, otherise remove XLINK placeholder from the template
	if ($xlink) {
		if ($html) {
			$template =~
			  s/XLINK/xmlns:xlink="http:\/\/www.w3.org\/1999\/xlink" xmlns:html="http:\/\/www.w3.org\/1999\/html" /;
		}
		else {
			# only xlink namepsace needed
			$template =~ s/XLINK/xmlns:xlink="http:\/\/www.w3.org\/1999\/xlink" /;
		}
	}
	elsif ($html) {

		# html namespace needed
		$template =~ s/XLINK/xmlns:html="http:\/\/www.w3.org\/1999\/html" /;
	}
	else {
		# no additional namespaces needed remove placeholder
		$template =~ s/XLINK//;
	}

	$template =~ s/COMMENT/$_[1]/;
	$template =~ s/TESTABLE_STATEMENT/$_[2]/;

	return $template;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 write_testfile
 
 Parameters  : contents, directory, test_name, suffix
 Returns     : the filename of the test file
 Description : writes the contents to a file in directory, the name of the file
			   is determined from test_name and suffix is added
 
=cut

sub write_testfile ($$$$) {
	my $contents = $_[0];
	chdir $_[1];
	my $filename = create_filename( $_[2], $_[3] );
	open( my $fh, ">:encoding(UTF-8)", $filename )
	  || die "can't open UTF-8 encoded $filename: $!";
	print $fh $contents;
	close $fh;
	return $filename;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 create_filename
 
 Parameters  : testname, suffix
 Returns     : filename (unused in cwd)
 Description : trys to use the filename + suffix, if it exists it adds integers to
			   filename until an unused filename is found. 
 
=cut

sub create_filename($$) {
	my $name     = create_filename_from_testname( $_[0] );
	my $iter     = 0;
	my $filename = $name . $_[1];

	# test whether the $filename is unused, if not add an integer to the end of the filename
	while ( -e $filename ) {
		$iter++;
		$filename = $name . $iter . $_[1];
	}

	return $filename;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 test_type
 
 Parameters  : wiki header line
 Returns     : test type 
 Description : pulls the test type by finding the word after Testable statements for.
 
=cut

sub test_type($) {
	if ( index( $_[0], 'not included in accessibility tree' ) > -1 ) {
		return 'not in accessibility tree';
	}
	my $ind = index( $_[0], 'for' );
	my $type = substr( $_[0], $ind + 3 );

	$ind = index( $type, "==" );
	if ( $ind > 0 ) {
		$type = substr( $type, 0, $ind );
	}
	$type =~ s/^\s+|\s+$//go;    # trim
	return $type;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 get_expected_table
 
 Parameters  : gets the wiki markup for the expected results table
 Returns     : wiki table 
 Description : finds the wiki table for expected results and returns it
 
=cut

sub get_expected_table($) {
	my $start = index( $_[0], '{|' );
	my $end = index( $_[0], '|}', $start ) + 1;
	return substr( $_[0], $start, ( $end - $start ) );
}

=head2 get_json_from_table
 
 Parameters  : wiki entry, indent
 Returns     : JSON for expected result
 Description : builds JSON for the expected results wiki table
 
=cut

sub get_json_from_table($$) {
	# this function uses indents and line returns to make the JSON human readable

	my $wikiTable = get_expected_table( $_[0] );
	my @rows = split( '\|-', $wikiTable ); # split of the wiki table into rows
	pop @rows;  # remove note row

	my $str = "{ ";			# JSON object string
	my $len = scalar @rows;
	my $indent = $_[1];		# used when an API has mulitple rows
	my $rowsInApi = 0;		# counter and flag used to indicate an API has more rows
	
	for ( my $i = 1 ; $i < $len ; $i++ ) {
		my $row = $rows[$i];	# the wiki table row
		my @entries = split( '\|', $row );    
		shift @entries; # first entry is an artifact of split and does not contain a property or value
		
		if (!$rowsInApi && $i > 1) {			
			$str .= ' ' x ($_[1] + 2);		# adds an indent to align APIs 
		}		
		
		if ($rowsInApi) {
			# not the first row for an API, indent includes the API name 
			$rowsInApi--;
			$str .= ' ' x $indent .  clean_quote( $entries[0] ) . ": " . clean_quote( $entries[1] );
			if (!$rowsInApi) {
				$str .= " }";
			}
		} elsif (scalar @entries == 2 ) {
			# API only has one value (ie not in tree), using a string instead of an object
			$str .= clean_quote( $entries[0] ) . ": " . clean_quote( $entries[1] );
		} elsif ( index( $row, 'rowspan' ) > -1 ) {
			# first row of a multiple row API
			$rowsInApi = get_rowspan($row);
			$rowsInApi--;
			shift @entries; # first entry is rowspan
						
			my $api = clean_quote($entries[0]);
			$str .= $api . ": { " . clean_quote( $entries[1] ) . ": " . clean_quote( $entries[2] ); 
			$indent = calculate_indent($api, 6, $_[1]); # indent includes the API name
		} elsif (scalar @entries == 3 ) {
			# API with a single row and a property/value pair
			my $api = clean_quote($entries[0]);
			$str .= $api . ": { " . clean_quote( $entries[1] ) . ": " . clean_quote( $entries[2] ) . " }";
		}
		
		# close the JSON object or add a comma
		$str = ($i == $len -1) ? $str . "\n". ' ' x $_[1] . "}\n" : $str . ",\n";		
	}

	return $str;
}

#/////////////////////////////////////////////////////////////////////////////

=head2 create_filename_from_testname
 
 Parameters  : testname
 Returns     : filename
 Description : converts a testname to a filename.
 
=cut

sub create_filename_from_testname($) {
	my $test_name = $_[0];

	# replace whitespace with single underscore
	$test_name =~ s/\s+_*/_/g;

	# remove any non-word characters
	$test_name =~ s/\W//g;

	# taking out word accessible to make names shorter
	$test_name =~ s/_*accessible//;

	# taking out word for to make names shorter
	$test_name =~ s/_+for_+/ /;

	# remove any trailing underscores
	$test_name =~ s/_+$//;

	# remove any leading underscores
	$test_name =~ s/^_+//;
	return $test_name;
}

=head2 read_testfile
 
 Parameters  : filename
 Returns     : hash of array of tests, key is test type (role, name, description)
 Description : gets the tests as a description [0] and a testable statement [1]
 
=cut

sub read_testfile ($) {
	open( FH, '<:encoding(UTF-8)', $_[0] ) or die "Could not open file '$_[0]' $!";
	my @lines = <FH>;
	close FH or die "Cannot close $_[0]: $!";

	my %alltests;
	my $tests;
	my $todos   = 0;
	my $counter = 0;
	for ( my $i = 0 ; $i < @lines ; $i++ ) {
		if ( $lines[$i] =~ /^\s*==\s+/ && $lines[$i] =~ /\s+==$/ ) {
			my $type     = test_type( $lines[$i] );
			my @testtype = ();
			$alltests{$type} = \@testtype;
			$tests           = \@testtype;
			$counter         = 0;
		}
		if ( $lines[$i] =~ /^;\s*\/\// ) {
			if ( $lines[ $i + 1 ] =~ /^\s*if/ ) {
				my $tlines = $lines[ $i + 1 ];
				my $index  = $i + 2;

				# gather up lines to the end of the expected results table
				while ( $index < @lines && index( $tlines, '|}' ) == -1 ) {
					$tlines .= $lines[$index];
					$index++;
				}

				my @test = ( $lines[$i], $tlines );
				@$tests[$counter] = \@test;
				$counter++;
				$i = $index;
			}
		}
	}
	return %alltests;
}

=head2 get_svg
 
 Parameters  : testable statement
 Returns     : svg  
 Description : gets the svg to be placed in the template
 
=cut

sub get_svg($) {
	
	my $svg = $_[0];

	# substring from the first < to the 'then' clause
	my $start = index( $_[0], '<' );
	my $then = index( $_[0], 'then' );
	my $end = rindex( $_[0], '>', $then ) + 1;
	return substr( $_[0], $start, ( $end - $start ) );
}

=head2 get_element_type
 
 Parameters  : svg
 Returns     : element (type) used in test  
 Description : returns the element type with the id="test"
 
=cut

sub get_element_type($) {
	my $start = get_element_open( $_[0] ) + 1;
	my $end = index( $_[0], " ", $start );
	return substr( $_[0], $start, $end - $start );
}

=head2 get_element_open
 
 Parameters  : svg
 Returns     : position of open angle bracket of the test element  
 Description : finds the postion of the open angle bracket of the test element
 
=cut

sub get_element_open($) {

	# finds id = test, uses double quotes, then single quotes and weakens the test if needed
	# find id="test"
	my $id = index( $_[0], 'id="test"' );
	if ( $id == -1 ) {

		# not found, try single quotes
		$id = index( $_[0], "id='test'" );
		if ( $id == -1 ) {

			# not found, try weaker test
			$id = index( $_[0], '"test"' );
			if ( $id == -1 ) {

				# not found, try single quotes
				$id = index( $_[0], "'test'" );
			}
		}
	}

	#	find the first open angle bracket before the id
	return rindex( $_[0], '<', $id );
}

=head2 calculate_indent

 Parameters  : key, bracket length, base indent length
 Returns     : indent length
 Description : calulates the indent length
 
=cut

sub calculate_indent($$$) {
	return length( $_[0] ) + $_[1] + $_[2];
}

=head2 get_rowspan
 
 Parameters  : row string
 Returns     : rowSpan 
 Description : returns rowspan value from the row string
 
=cut

sub get_rowspan($) {
	my $eqInd = index( $_[0], 'rowspan=' ) + 8; 
	if ($eqInd < 0) {
		return 0;
	}
	my $qtInd = index($_[0], '"', $eqInd) + 1;
	my $qt2Ind = index($_[0], '"', $qtInd);
	my $str = substr( $_[0], $qtInd, $qt2Ind - $qtInd);
	$str =~ s/^\s+//o;    # trims beginning whitespace
	return $str;
}
=head2 escape_quotes
 
 Parameters  : string to escape
 Returns     : modified string
 Description : adds escapes (\) to strings on double quotes.
 
=cut

sub escape_quotes($) {		
	return join('\"', split('"', $_[0]));
}

=head2 file_to_json
 
 Parameters  : type, file string
 Returns     : json string for object
 Description : converts a type and a file string into a JSON object string.
 
=cut
sub file_to_json($$) {
	 my $json = "    { \"type\":\"". $_[0]. "\", \n    \"test\":\n" ;
	$json .= "\"". join('\"', split('"', $_[1])) . "\"\n }";
}

=head2 clean_for_output
 
 Parameters  : string to clean
 Returns     : cleaned string
 Description : cleans a string to be put into a csv file.
 
=cut

sub clean_for_output($) {
	$string = $_[0];

	# replace comma with a space
	$string =~ s/,+/ /og;

	# remove an equal sign
	$string =~ s/\s*=\s*//og;

	# remove quotes
	$string =~ s/\"*\'*//og;

	# remove a newline
	$string =~ s/\n*//og;

	# remove excess whitespace
	$string =~ s/\s+/ /og;
	$string =~ s/^\s//o;
	$string =~ s/\s$//o;
	return $string;
}

=head2 clean_quote
 
 Parameters  : string to clean and quote
 Returns     : cleaned quoted string
 Description : cleans and quotes a string for JSON
 
=cut

sub clean_quote($) {
	$string = $_[0];

	#	# replace comma with a space
	#	$string =~ s/,+/ /og;
	#
	#	# remove an equal sign
	#	$string =~ s/\s*=\s*//og;
	#
	#	# remove quotes
	#	$string =~ s/\"*\'*//og;

	# remove a newline
	$string =~ s/\n*//og;

	# remove excess whitespace
	$string =~ s/\s+/ /og;
	$string =~ s/^\s//o;
	$string =~ s/\s$//o;

	return '"' . $string . '"';
}

=head2 guess_section
 
 Parameters  : type
 Returns     : section URI
 Description : guesses the section of the spec for the test.
 
=cut

sub guess_section($) {
	my $var = $_[0];
	if ( $var == NAME || $var == DESCRIPTION ) {
		return "https://rawgit.com/w3c/aria/master/svg-aam/svg-aam.html#mapping_additional_nd";
	}

	if ( $var == ROLE_NONE || $var == ROLE_PRESENTATION ) {
		return "https://rawgit.com/w3c/aria/master/svg-aam/svg-aam.html#exclude_elements";
	}

	if ( $var == ROLE ) {
		return "https://rawgit.com/w3c/aria/master/svg-aam/svg-aam.html#mapping_role_table";
	}
	if ( $var == NOT_IN_TREE ) {
		return "https://rawgit.com/w3c/aria/master/svg-aam/svg-aam.html#mapping_role_table";
	}

	return '';
}

=head2 get_type_enum
 
 Parameters  : type string
 Returns     : type enum
 Description : returns the type enum 
 
=cut

sub get_type_enum($) {
	my $var = lc( $_[0] );
	$var =~ s/^\s+|\s+$//go;    #trim whitespace

	if ( $var eq "not in accessibility tree" ) {
		return NOT_IN_TREE;
	}

	if ( $var eq "name" ) {
		return NAME;
	}

	if ( $var eq "description" ) {
		return DESCRIPTION;
	}

	if ( $var eq "role is none" ) {
		return ROLE_NONE;
	}

	if ( $var eq "role is presentation" ) {
		return ROLE_PRESENTATION;
	}

	if ( index( $var, "role" ) > -1 ) {
		return ROLE;
	}

	# language tests are often more wordy
	if ( index( $var, "name" ) > -1 ) {
		return NAME;
	}

	if ( index( $var, "description" ) > -1 ) {
		return DESCRIPTION;
	}

	return 0;
}

=head2 add_json_test_object
 
 Parameters  : filehandle, type, element, comment, html, svg, expected result as JSON
 Returns     :  none
 Description : writes a json object for a test to the db file
 
=cut

sub add_json_test_object($$$$$$$) {
	my $fh = shift;

	my $json = "\n {\n  \"testing\":\"";
	$json .= clean_for_output( $_[0] );
	$json .= "\",\n  \"element\":\"";
	$json .= clean_for_output( $_[1] );
	$json .= "\",\n  \"comment\":\"";
	$json .= clean_for_output( $_[2] );
	$json .= "\",\n  \"expected\": ";

	if ( $_[5] ) {    # If expected result is empty, it is printed out in the main program as a bad file
		$json .= $_[5];
	}
	
	$json .= "  ,\n  \"files\": [\n";
	$json .= file_to_json("html", $_[3]) . ",\n";
	$json .= file_to_json("svg", $_[4]) . "\n ]";
	$json .= "}";
	
	print $fh $json;
}

# main program
# convert wiki markup with tables into JSON for test harness
# when switching groups of tests a line will begin and end with ==
# a test starts with a comment describing the test ; //  (whitespace may vary)
# expected result starts with   then  and has a wiki table 

# directories and file names
# where the wiki markup is - the wiki markup copied into this file
#my $wiki_markup_filename = 'c:\\bob\\wikitable.txt';
my $wiki_markup_filename = 'c:\\bob\\wikitableLang.txt';


#directories to write test files to and JSON files
$test_dir = "c:\\test\\";

# database filenames
my $json_filename     = $test_dir . "svghtml.json";


# create the test directories if they don't exist
mkdir($test_dir) unless ( -d $test_dir );


# alltests is a hash with all the tests from the wiki markup
my %alltests = read_testfile($wiki_markup_filename);

# template for HTML test files COMMENT, XLINK and TESTABLE_STATEMENT are replaced 
my $html_template = '<!DOCTYPE html>
<html>
   <head>
      <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
      <title> COMMENT </title>
   </head>
   <body>
   <svg xmlns="http://www.w3.org/2000/svg" XLINK >
     TESTABLE_STATEMENT
   </svg>
   </body>
</html>';

# template for SVG test files XLINK and TESTABLE_STATEMENT are replaced
my $svg_template = '<svg xmlns="http://www.w3.org/2000/svg" XLINK >
	TESTABLE_STATEMENT
</svg>';

#open the json files
open( my $json_html, ">:encoding(UTF-8)", $json_filename )
  || die "can't open UTF-8 encoded $json_filename: $!";

# start the json file (open an array)
print $json_html "[";

my $counter = 0;

# main loop, building test files and JSON files
foreach my $key ( keys %alltests ) {
	my $tests = $alltests{$key};

	# simplify the test type want to keep track of - description, name, role, not in accessiblity tree and roles NONE, PRESENTATION
	my $type = $key;
	$type =~ s/calculation//o;
	$type =~ s/Role/role/o;
	my $testType = get_type_enum($type);
	my $section  = guess_section($testType);
	my $typeCount = 0;
	
	# make the testfiles and populate the database files
	foreach my $test ( @{ $alltests{$key} } ) {

		my $svg     = get_svg( $$test[1] );
		my $element = get_element_type($svg);
		my $comment = $$test[0];
		$comment =~ s/;\s*\/\///;

		# make the HTML test file
		my $h = populate_test_template( $html_template, $comment, $svg );

		# make the SVG test file
		my $s = populate_test_template( $svg_template, $comment, $svg );

		# add the test to the JSON files
		if ( $counter > 0 ) {

			# add a comma before adding the new test
			print $json_html ',';
		}

		my $expectedJson = get_json_from_table( $$test[1], 14 );
		if ( !$expectedJson ) {
			print $svg_filename . " failed to create a good test   \n";
		}

		add_json_test_object( $json_html, $type, $element, $comment, $h, $s, $expectedJson );
		$counter++;
		$typeCount++;
	}

	print $typeCount. " tests for " . $type . "\n";

}

# ending the process, closing files
print $counter. " tests were written \n";

# finish and close JSON database files
print $json_html "\n]";
close $json_html;


print 'program exited normally';
